@@ -1,4 +1,5 @@
 use std::collections::{HashMap, HashSet};
+use std::iter::once;
 use std::marker::PhantomData;
 use std::ops::{Deref, DerefMut};
 
@@ -8,8 +9,8 @@ use quote::{format_ident, quote, ToTokens};
 use syn::parse::{Parse, ParseStream};
 use syn::punctuated::Punctuated;
 use syn::{
-    parenthesized, token, Attribute, Data, DataEnum, DeriveInput, Expr, Field, Fields, Generics,
-    Ident, Index, Token, Type,
+    bracketed, parenthesized, token, Attribute, Data, DataEnum, DeriveInput, Expr, Field, Fields,
+    Generics, Ident, Index, Token, Type, WhereClause,
 };
 
 use easy_proc::{find_attr, parse_attribute_list, ArgumentList};
@@ -23,6 +24,8 @@ pub struct AccountArgumentAttribute {
     #[allow(dead_code)]
     #[argument(attr_ident)]
     attr_ident: Ident,
+    account_info: Type,
+    generics: Option<AdditionalGenerics>,
     // TODO: Use this with enum derivation
     #[allow(dead_code)]
     #[argument(default = syn::parse_str("u64").unwrap())]
@@ -35,23 +38,15 @@ pub struct AccountArgumentAttribute {
 impl AccountArgumentAttribute {
     const IDENT: &'static str = "account_argument";
 }
-impl Default for AccountArgumentAttribute {
-    fn default() -> Self {
-        Self {
-            attr_ident: Ident::new("__does_not_exist__", Span::call_site()),
-            enum_discriminant_type: syn::parse_str("u64").unwrap(),
-            no_from: false,
-            no_validate: false,
-        }
-    }
-}
 
 #[derive(ArgumentList)]
 pub struct FromAttribute {
     #[argument(attr_ident)]
     attr_ident: Ident,
     id: Option<Ident>,
+    #[argument(default)]
     data: NamedTupple,
+    generics: Option<AdditionalGenerics>,
     // TODO: Use this for enum derivation
     #[allow(dead_code)]
     enum_discriminant: Option<Expr>,
@@ -82,6 +77,7 @@ impl Default for FromAttribute {
             attr_ident: Ident::new("__does_not_exist__", Span::call_site()),
             id: None,
             data: NamedTupple::default(),
+            generics: None,
             enum_discriminant: None,
             log_level: LogLevel::default(),
         }
@@ -93,7 +89,9 @@ pub struct ValidateAttribute {
     #[argument(attr_ident)]
     attr_ident: Ident,
     id: Option<Ident>,
+    #[argument(default)]
     data: NamedTupple,
+    generics: Option<AdditionalGenerics>,
     // TODO: add logging
     #[allow(dead_code)]
     #[argument(default)]
@@ -121,22 +119,23 @@ impl Default for ValidateAttribute {
             attr_ident: Ident::new("__does_not_exist__", Span::call_site()),
             id: None,
             data: NamedTupple::default(),
+            generics: None,
             log_level: LogLevel::default(),
         }
     }
 }
 
 #[derive(ArgumentList, Debug)]
-struct ArgumentFromFieldAttribute {
+struct FromFieldAttribute {
     #[argument(attr_ident)]
     attr_ident: Ident,
     id: Option<Ident>,
     data: Option<Expr>,
 }
-impl ArgumentFromFieldAttribute {
+impl FromFieldAttribute {
     const IDENT: &'static str = "from";
 }
-impl IdAttr for ArgumentFromFieldAttribute {
+impl IdAttr for FromFieldAttribute {
     fn id(&self) -> Option<&Ident> {
         self.id.as_ref()
     }
@@ -145,7 +144,7 @@ impl IdAttr for ArgumentFromFieldAttribute {
         &self.attr_ident
     }
 }
-impl Default for ArgumentFromFieldAttribute {
+impl Default for FromFieldAttribute {
     fn default() -> Self {
         Self {
             attr_ident: Ident::new("__invalid_identifier__", Span::call_site()),
@@ -156,7 +155,7 @@ impl Default for ArgumentFromFieldAttribute {
 }
 
 #[derive(ArgumentList, Debug, Clone)]
-struct ArgumentValidateFieldAttribute {
+struct ValidateFieldAttribute {
     #[argument(attr_ident)]
     attr_ident: Ident,
     id: Option<Ident>,
@@ -170,10 +169,10 @@ struct ArgumentValidateFieldAttribute {
     #[argument(custom)]
     key: Option<IndexesValue<Expr, UnitDefault>>,
 }
-impl ArgumentValidateFieldAttribute {
+impl ValidateFieldAttribute {
     const IDENT: &'static str = "validate";
 }
-impl IdAttr for ArgumentValidateFieldAttribute {
+impl IdAttr for ValidateFieldAttribute {
     fn id(&self) -> Option<&Ident> {
         self.id.as_ref()
     }
@@ -182,7 +181,7 @@ impl IdAttr for ArgumentValidateFieldAttribute {
         &self.attr_ident
     }
 }
-impl Default for ArgumentValidateFieldAttribute {
+impl Default for ValidateFieldAttribute {
     fn default() -> Self {
         Self {
             attr_ident: Ident::new("__invalid_identifier__", Span::call_site()),
@@ -196,6 +195,34 @@ impl Default for ArgumentValidateFieldAttribute {
     }
 }
 
+#[derive(Clone, Debug)]
+struct AdditionalGenerics {
+    bracket: token::Bracket,
+    generics: Generics,
+    where_clause: Option<WhereClause>,
+}
+impl Parse for AdditionalGenerics {
+    fn parse(input: ParseStream) -> syn::Result<Self> {
+        let content;
+        let bracket = bracketed!(content in input);
+        let generics = content.parse()?;
+        let where_clause = content.parse()?;
+        Ok(Self {
+            bracket,
+            generics,
+            where_clause,
+        })
+    }
+}
+impl ToTokens for AdditionalGenerics {
+    fn to_tokens(&self, tokens: &mut TokenStream) {
+        self.bracket.surround(tokens, |tokens| {
+            self.generics.to_tokens(tokens);
+            self.where_clause.to_tokens(tokens);
+        });
+    }
+}
+
 trait IdAttr: ArgumentList {
     fn id(&self) -> Option<&Ident>;
     fn attr_ident(&self) -> &Ident;
@@ -285,17 +312,24 @@ impl Parse for AccountArgumentDerive {
     fn parse(input: ParseStream) -> syn::Result<Self> {
         let from_attribute_ident = format_ident!("{}", FromAttribute::IDENT);
         let validate_attribute_ident = format_ident!("{}", ValidateAttribute::IDENT);
-        let argument_from_field_attr_ident = format_ident!("{}", ArgumentFromFieldAttribute::IDENT);
-        let argument_validate_field_attr_ident =
-            format_ident!("{}", ArgumentValidateFieldAttribute::IDENT);
+        let argument_from_field_attr_ident = format_ident!("{}", FromFieldAttribute::IDENT);
+        let argument_validate_field_attr_ident = format_ident!("{}", ValidateFieldAttribute::IDENT);
         let derive_input: DeriveInput = input.parse()?;
 
         let account_argument_attribute = find_attr(
             derive_input.attrs.iter(),
             &format_ident!("{}", AccountArgumentAttribute::IDENT),
         )
-        .map(AccountArgumentAttribute::parse_arguments)
-        .unwrap_or_default();
+        .map_or_else(
+            || {
+                abort!(
+                    derive_input.ident,
+                    "Missing `{}` attribute",
+                    AccountArgumentAttribute::IDENT
+                )
+            },
+            AccountArgumentAttribute::parse_arguments,
+        );
 
         let mut from_attributes =
             FromAttribute::read_all(&from_attribute_ident, derive_input.attrs.iter());
@@ -331,8 +365,14 @@ impl AccountArgumentDerive {
             TokenStream::new()
         } else {
             let from_accounts = self.from_attributes.into_iter().map(|(id, attr)| {
-                self.derive_type
-                    .from_accounts(&self.ident, &self.generics, id, attr)
+                self.derive_type.from_accounts(
+                    &self.ident,
+                    &self.generics,
+                    self.account_argument_attribute.generics.as_ref(),
+                    &id,
+                    &attr,
+                    &self.account_argument_attribute.account_info,
+                )
             });
             quote! { #(#from_accounts)* }
         };
@@ -341,8 +381,13 @@ impl AccountArgumentDerive {
             TokenStream::new()
         } else {
             let validate_argument = self.validate_attributes.into_iter().map(|(id, attr)| {
-                self.derive_type
-                    .validate_argument(&self.ident, &self.generics, id, attr)
+                self.derive_type.validate_argument(
+                    &self.ident,
+                    &self.generics,
+                    self.account_argument_attribute.generics.as_ref(),
+                    &id,
+                    &attr,
+                )
             });
             quote! { #(#validate_argument)* }
         };
@@ -357,16 +402,25 @@ impl AccountArgumentDerive {
     fn account_argument(&self) -> TokenStream {
         let crate_name = get_crate_name();
         let ident = &self.ident;
-        let (impl_gen, ty_gen, where_clause) = self.generics.split_for_impl();
+
+        let (impl_gen, ty_gen, where_clause) = combine_generics(
+            &self.generics,
+            once(self.account_argument_attribute.generics.as_ref()),
+        );
+
         let write_back = self.derive_type.write_back();
         let add_keys = self.derive_type.add_keys();
+        let account_info = &self.account_argument_attribute.account_info;
 
         quote! {
             #[automatically_derived]
+            #[allow(clippy::type_repetition_in_bounds)]
             impl #impl_gen #crate_name::account_argument::AccountArgument for #ident #ty_gen #where_clause {
+                type AccountInfo = #account_info;
+
                 fn write_back(
                     self,
-                    program_id: &'static #crate_name::Pubkey,
+                    program_id: &#crate_name::Pubkey,
                 ) -> #crate_name::CruiserResult<()>{
                     #write_back
                     Ok(())
@@ -374,7 +428,7 @@ impl AccountArgumentDerive {
 
                 fn add_keys(
                     &self,
-                    mut add__: impl ::core::ops::FnMut(&'static #crate_name::solana_program::pubkey::Pubkey) -> #crate_name::CruiserResult<()>
+                    mut add__: impl ::core::ops::FnMut(#crate_name::solana_program::pubkey::Pubkey) -> #crate_name::CruiserResult<()>
                 ) -> #crate_name::CruiserResult<()>{
                     #add_keys
                     Ok(())
@@ -384,6 +438,39 @@ impl AccountArgumentDerive {
     }
 }
 
+/// (`impl_gen`, `ty_gen`, `where_clause`)
+#[must_use]
+fn combine_generics<'a>(
+    generics: &Generics,
+    other_generics: impl IntoIterator<Item = Option<&'a AdditionalGenerics>>,
+) -> (TokenStream, TokenStream, TokenStream) {
+    let type_params = generics.type_params();
+    let mut generics = generics.clone();
+    for other_generics in other_generics.into_iter().flatten() {
+        generics
+            .params
+            .extend(other_generics.generics.params.iter().cloned());
+        for where_clause in [
+            &other_generics.generics.where_clause,
+            &other_generics.where_clause,
+        ]
+        .into_iter()
+        .flatten()
+        {
+            generics
+                .make_where_clause()
+                .predicates
+                .extend(where_clause.predicates.iter().cloned());
+        }
+    }
+    let (impl_gen, _, where_clause) = generics.split_for_impl();
+    (
+        quote! { #impl_gen },
+        quote! { <#(#type_params,)*> },
+        quote! { #where_clause },
+    )
+}
+
 enum AccountArgumentDeriveType {
     Enum(AccountArgumentDeriveEnum),
     Struct(AccountArgumentDeriveStruct),
@@ -405,15 +492,15 @@ impl AccountArgumentDeriveType {
                     argument_validate_field_attr_ident,
                     from_ids,
                     validate_ids,
-                )?))
+                )))
             }
             Data::Enum(data_enum) => Ok(Self::Enum(AccountArgumentDeriveEnum::from_enum(
                 data_enum,
                 argument_from_field_attr_ident,
                 argument_validate_field_attr_ident,
-                from_ids,
-                validate_ids,
-            )?)),
+                &from_ids,
+                &validate_ids,
+            ))),
             Data::Union(union) => {
                 abort!(
                     union.union_token.span.join(ident.span()).unwrap(),
@@ -444,11 +531,16 @@ impl AccountArgumentDeriveType {
         &self,
         ident: &Ident,
         generics: &Generics,
-        id: String,
-        attr: FromAttribute,
+        argument_generics: Option<&AdditionalGenerics>,
+        id: &str,
+        attr: &FromAttribute,
+        account_info: &Type,
     ) -> TokenStream {
         let crate_name = get_crate_name();
-        let (impl_gen, ty_gen, where_clause) = generics.split_for_impl();
+
+        let (impl_gen, ty_gen, where_clause) =
+            combine_generics(generics, [attr.generics.as_ref(), argument_generics]);
+
         let ty_accessors = attr.to_type(&quote! { __arg });
         let program_id = quote! { program_id };
         let infos = quote! { __infos };
@@ -457,15 +549,16 @@ impl AccountArgumentDeriveType {
             let inner = match self {
                 AccountArgumentDeriveType::Enum(_) => todo!(),
                 AccountArgumentDeriveType::Struct(data) => {
-                    data.from_accounts(&id, &program_id, &infos)
+                    data.from_accounts(id, &program_id, &infos)
                 }
             };
             out.push(quote! {
                 #[automatically_derived]
+                #[allow(clippy::type_repetition_in_bounds)]
                 impl #impl_gen #crate_name::account_argument::FromAccounts<#ty> for #ident #ty_gen #where_clause{
                     fn from_accounts(
-                        program_id: &'static #crate_name::Pubkey,
-                        __infos: &mut impl #crate_name::account_argument::AccountInfoIterator,
+                        program_id: &#crate_name::Pubkey,
+                        __infos: &mut impl #crate_name::account_argument::AccountInfoIterator<Item = #account_info>,
                         __arg: #ty,
                     ) -> #crate_name::CruiserResult<Self>{
                         #(#accessors)*
@@ -488,11 +581,15 @@ impl AccountArgumentDeriveType {
         &self,
         ident: &Ident,
         generics: &Generics,
-        id: String,
-        attr: ValidateAttribute,
+        argument_generics: Option<&AdditionalGenerics>,
+        id: &str,
+        attr: &ValidateAttribute,
     ) -> TokenStream {
         let crate_name = get_crate_name();
-        let (impl_gen, ty_gen, where_clause) = generics.split_for_impl();
+
+        let (impl_gen, ty_gen, where_clause) =
+            combine_generics(generics, [attr.generics.as_ref(), argument_generics]);
+
         let ty_accessors = attr.to_type(&quote! { __arg });
         let program_id = quote! { program_id };
         let mut out = Vec::with_capacity(ty_accessors.len());
@@ -500,13 +597,14 @@ impl AccountArgumentDeriveType {
             let inner = match self {
                 AccountArgumentDeriveType::Enum(_) => todo!(),
                 AccountArgumentDeriveType::Struct(data) => {
-                    data.validate_argument(&id, &program_id, &quote! { self. })
+                    data.validate_argument(id, &program_id, &quote! { self. })
                 }
             };
             out.push(quote! {
                 #[automatically_derived]
+                #[allow(clippy::type_repetition_in_bounds)]
                 impl #impl_gen #crate_name::account_argument::ValidateArgument<#ty> for #ident #ty_gen #where_clause{
-                    fn validate(&mut self, program_id: &'static #crate_name::Pubkey, __arg: #ty) -> #crate_name::CruiserResult<()>{
+                    fn validate(&mut self, program_id: &#crate_name::Pubkey, __arg: #ty) -> #crate_name::CruiserResult<()>{
                         #(#accessors)*
                         #inner
                         ::std::result::Result::Ok(())
@@ -527,9 +625,9 @@ impl AccountArgumentDeriveEnum {
         value: DataEnum,
         argument_from_field_attr_ident: &Ident,
         argument_validate_field_attr_ident: &Ident,
-        from_ids: HashSet<String>,
-        validate_ids: HashSet<String>,
-    ) -> syn::Result<Self> {
+        from_ids: &HashSet<String>,
+        validate_ids: &HashSet<String>,
+    ) -> Self {
         let mut variants = Vec::with_capacity(value.variants.len());
         for variant in value.variants {
             variants.push(AccountArgumentEnumVariant {
@@ -540,11 +638,11 @@ impl AccountArgumentDeriveEnum {
                     argument_validate_field_attr_ident,
                     from_ids.clone(),
                     validate_ids.clone(),
-                )?,
+                ),
                 discriminant: variant.discriminant.map(|(_, discriminant)| discriminant),
-            })
+            });
         }
-        Ok(Self(variants))
+        Self(variants)
     }
 
     fn write_back(&self) -> TokenStream {
@@ -665,9 +763,9 @@ impl AccountArgumentDeriveStruct {
         argument_validate_field_attr_ident: &Ident,
         from_ids: HashSet<String>,
         validate_ids: HashSet<String>,
-    ) -> Result<Self, syn::Error> {
+    ) -> Self {
         match value {
-            Fields::Named(named) => Ok(Self::Named(
+            Fields::Named(named) => Self::Named(
                 Self::from_named(
                     named.named.into_iter(),
                     argument_from_field_attr_ident,
@@ -676,8 +774,8 @@ impl AccountArgumentDeriveStruct {
                     validate_ids,
                 )
                 .collect(),
-            )),
-            Fields::Unnamed(unnamed) => Ok(Self::Unnamed(
+            ),
+            Fields::Unnamed(unnamed) => Self::Unnamed(
                 Self::from_unnamed(
                     unnamed.unnamed.into_iter(),
                     argument_from_field_attr_ident,
@@ -686,8 +784,8 @@ impl AccountArgumentDeriveStruct {
                     validate_ids,
                 )
                 .collect(),
-            )),
-            Fields::Unit => Ok(Self::Unit),
+            ),
+            Fields::Unit => Self::Unit,
         }
     }
 
@@ -720,16 +818,14 @@ impl AccountArgumentDeriveStruct {
         validate_ids: HashSet<String>,
     ) -> impl Iterator<Item = UnnamedField> + 'a {
         value.map(move |field| {
-            let from_attrs = ArgumentFromFieldAttribute::read_all(
-                argument_from_field_attr_ident,
-                field.attrs.iter(),
-            );
-            let validate_attrs = ArgumentValidateFieldAttribute::read_all(
+            let from_attrs =
+                FromFieldAttribute::read_all(argument_from_field_attr_ident, field.attrs.iter());
+            let validate_attrs = ValidateFieldAttribute::read_all(
                 argument_validate_field_attr_ident,
                 field.attrs.iter(),
             );
 
-            for (key, value) in from_attrs.iter() {
+            for (key, value) in &from_attrs {
                 if !from_ids.contains(key) {
                     match &value.id {
                         Some(id) => abort!(id, "Unknown id `{}`", id),
@@ -737,7 +833,7 @@ impl AccountArgumentDeriveStruct {
                     }
                 }
             }
-            for (key, value) in validate_attrs.iter() {
+            for (key, value) in &validate_attrs {
                 if !validate_ids.contains(key) {
                     match &value.id {
                         Some(id) => abort!(id, "Unknown id `{}`", id),
@@ -961,8 +1057,8 @@ impl DerefMut for NamedField {
 
 #[derive(Debug)]
 struct UnnamedField {
-    from_attrs: HashMap<String, ArgumentFromFieldAttribute>,
-    validate_attrs: HashMap<String, ArgumentValidateFieldAttribute>,
+    from_attrs: HashMap<String, FromFieldAttribute>,
+    validate_attrs: HashMap<String, ValidateFieldAttribute>,
     ty: Type,
 }
 impl UnnamedField {
@@ -996,7 +1092,7 @@ impl UnnamedField {
             .get(id)
             .and_then(|attr| attr.data.clone())
             .unwrap_or_else(|| syn::parse_str("()").unwrap());
-        quote! { #crate_name::account_argument::FromAccounts::from_accounts(#program_id, #infos, #expr)? }
+        quote! { #crate_name::account_argument::FromAccounts::<_>::from_accounts(#program_id, #infos, #expr)? }
     }
 
     fn validate_argument(
@@ -1028,7 +1124,7 @@ impl UnnamedField {
         });
 
         quote! {
-            #crate_name::account_argument::ValidateArgument::validate(&mut #accessor, #program_id, #validate)?;
+            #crate_name::account_argument::ValidateArgument::<_>::validate(&mut #accessor, #program_id, #validate)?;
             #(#signer)*
             #(#writable)*
             #(#owner)*
